@@ -50,7 +50,7 @@ let read_member_info pack_path file = (
       if info.ui_name <> name
       then raise(Error(Illegal_renaming(name, file, info.ui_name)));
       if info.ui_symbol <>
-         (Compilenv.current_unit_infos()).ui_symbol ^ "__" ^ info.ui_name
+         (Compilenv.current_unit_infos()).ui_symbol ^ "." ^ info.ui_name
       then raise(Error(Wrong_for_pack(file, pack_path)));
       Asmlink.check_consistency file info crc;
       Compilenv.cache_unit_info info;
@@ -232,6 +232,7 @@ let build_package_cmx members cmxfile =
       ui_force_link =
           List.exists (fun info -> info.ui_force_link) units;
       ui_export_info;
+      ui_for_pack = None;
     } in
   Compilenv.write_unit_info pkg_infos cmxfile
 
