@@ -336,14 +336,12 @@ static void wo_memmove (volatile value* const dst,
     if (dst < src) {
       /* copy ascending */
       for (i = 0; i < nvals; i++)
-        atomic_store_explicit(&((atomic_value*)dst)[i], src[i],
-                              memory_order_release);
+        atomic_store_release(&((atomic_value*)dst)[i], src[i]);
 
     } else {
       /* copy descending */
       for (i = nvals; i > 0; i--)
-        atomic_store_explicit(&((atomic_value*)dst)[i-1], src[i-1],
-                              memory_order_release);
+        atomic_store_release(&((atomic_value*)dst)[i-1], src[i-1]);
     }
   }
 }
@@ -586,7 +584,7 @@ CAMLprim value caml_array_fill(value array,
       *fp = val;
       if (Is_block(old)) {
         if (Is_young(old)) continue;
-        caml_darken(NULL, old, NULL);
+        caml_darken(Caml_state, old, NULL);
       }
       if (is_val_young_block)
         Ref_table_add(&Caml_state->minor_tables->major_ref, fp);
