@@ -274,7 +274,7 @@ static void remove_from_stw_domains(dom_internal* dom) {
   stw_domains.domains[stw_domains.participating_domains] = dom;
 }
 
-static dom_internal* next_free_domain() {
+static dom_internal* next_free_domain(void) {
   if (stw_domains.participating_domains == Max_domains)
     return NULL;
 
@@ -294,22 +294,22 @@ CAMLexport caml_domain_state* caml_get_domain_state(void)
 
 Caml_inline void interrupt_domain(struct interruptor* s)
 {
-  atomic_store_rel(s->interrupt_word, (uintnat)(-1));
+  atomic_store_release(s->interrupt_word, (uintnat)(-1));
 }
 
 int caml_incoming_interrupts_queued(void)
 {
-  return atomic_load_acq(&domain_self->interruptor.interrupt_pending);
+  return atomic_load_acquire(&domain_self->interruptor.interrupt_pending);
 }
 
 /* must NOT be called with s->lock held */
 static void stw_handler(caml_domain_state* domain);
 static uintnat handle_incoming(struct interruptor* s)
 {
-  uintnat handled = atomic_load_acq(&s->interrupt_pending);
+  uintnat handled = atomic_load_acquire(&s->interrupt_pending);
   CAMLassert (s->running);
   if (handled) {
-    atomic_store_rel(&s->interrupt_pending, 0);
+    atomic_store_release(&s->interrupt_pending, 0);
 
     stw_handler(domain_self->state);
   }
@@ -330,7 +330,7 @@ void caml_handle_incoming_interrupts(void)
 int caml_send_interrupt(struct interruptor* target)
 {
   /* signal that there is an interrupt pending */
-  atomic_store_rel(&target->interrupt_pending, 1);
+  atomic_store_release(&target->interrupt_pending, 1);
 
   /* Signal the condition variable, in case the target is
      itself waiting for an interrupt to be processed elsewhere */
@@ -349,7 +349,7 @@ static void caml_wait_interrupt_serviced(struct interruptor* target)
 
   /* Often, interrupt handlers are fast, so spin for a bit before waiting */
   for (i=0; i<1000; i++) {
-    if (!atomic_load_acq(&target->interrupt_pending)) {
+    if (!atomic_load_acquire(&target->interrupt_pending)) {
       return;
     }
     cpu_relax();
@@ -357,7 +357,7 @@ static void caml_wait_interrupt_serviced(struct interruptor* target)
 
   {
     SPIN_WAIT {
-      if (!atomic_load_acq(&target->interrupt_pending))
+      if (!atomic_load_acquire(&target->interrupt_pending))
         return;
     }
   }
@@ -411,7 +411,7 @@ asize_t caml_norm_minor_heap_size (intnat wsize)
   domain, so they need no synchronization.
 */
 
-Caml_inline void check_minor_heap() {
+Caml_inline void check_minor_heap(void) {
   caml_domain_state* domain_state = Caml_state;
   CAMLassert(domain_state->young_ptr == domain_state->young_end);
 
@@ -433,7 +433,7 @@ Caml_inline void check_minor_heap() {
       && domain_state->young_end <= (value*)domain_self->minor_heap_area_end));
 }
 
-static void free_minor_heap() {
+static void free_minor_heap(void) {
   caml_domain_state* domain_state = Caml_state;
 
   caml_gc_log ("trying to free old minor heap: %"
@@ -449,10 +449,11 @@ static void free_minor_heap() {
       (void*)domain_self->minor_heap_area_start,
       Bsize_wsize(domain_state->minor_heap_wsz));
 
-  domain_state->young_start =
-    domain_state->young_end =
-    domain_state->young_ptr = NULL;
-  atomic_store_rel(&domain_state->young_limit,
+  domain_state->young_start   = NULL;
+  domain_state->young_end     = NULL;
+  domain_state->young_ptr     = NULL;
+  domain_state->young_trigger = NULL;
+  atomic_store_release(&domain_state->young_limit,
                    (uintnat) domain_state->young_start);
 }
 
@@ -490,6 +491,10 @@ static int allocate_minor_heap(asize_t wsize) {
   domain_state->young_end =
       (value*)(domain_self->minor_heap_area_start + Bsize_wsize(wsize));
   domain_state->young_ptr = domain_state->young_end;
+  /* Trigger a GC poll when half of the minor heap is filled. At that point, a
+   * major slice is scheduled. */
+  domain_state->young_trigger = domain_state->young_start
+         + (domain_state->young_end - domain_state->young_start) / 2;
   caml_reset_young_limit(domain_state);
 
   check_minor_heap();
@@ -540,7 +545,7 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
   caml_plat_lock(&all_domains_lock);
 
   /* Wait until any in-progress STW sections end. */
-  while (atomic_load_acq(&stw_leader)) {
+  while (atomic_load_acquire(&stw_leader)) {
     /* [caml_plat_wait] releases [all_domains_lock] until the current
        STW section ends, and then takes the lock again. */
     caml_plat_wait(&all_domains_cond);
@@ -597,11 +602,14 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
   domain_state->dependent_size = 0;
   domain_state->dependent_allocated = 0;
 
+  domain_state->major_work_done_between_slices = 0;
+
   /* the minor heap will be initialized by
      [caml_reallocate_minor_heap] below. */
-  domain_state->young_start =
-    domain_state->young_end =
-    domain_state->young_ptr = NULL;
+  domain_state->young_start = NULL;
+  domain_state->young_end = NULL;
+  domain_state->young_ptr = NULL;
+  domain_state->young_trigger = NULL;
 
   domain_state->minor_tables = caml_alloc_minor_tables();
   if(domain_state->minor_tables == NULL) {
@@ -662,6 +670,7 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
 
   domain_state->requested_major_slice = 0;
   domain_state->requested_minor_gc = 0;
+  domain_state->major_slice_epoch = 0;
   domain_state->requested_external_interrupt = 0;
 
   domain_state->parser_trace = 0;
@@ -712,7 +721,7 @@ CAMLexport void caml_reset_domain_lock(void)
 
 /* minor heap initialization and resizing */
 
-static void reserve_minor_heaps() {
+static void reserve_minor_heaps(void) {
   void* heaps_base;
   uintnat minor_heap_reservation_bsize;
   uintnat minor_heap_max_bsz;
@@ -724,7 +733,7 @@ static void reserve_minor_heaps() {
   minor_heap_reservation_bsize = minor_heap_max_bsz * Max_domains;
 
   /* reserve memory space for minor heaps */
-  heaps_base = caml_mem_map(minor_heap_reservation_bsize, caml_sys_pagesize,
+  heaps_base = caml_mem_map(minor_heap_reservation_bsize, caml_plat_pagesize,
                 1 /* reserve_only */);
   if (heaps_base == NULL)
     caml_fatal_error("Not enough heap memory to reserve minor heaps");
@@ -749,7 +758,7 @@ static void reserve_minor_heaps() {
   }
 }
 
-static void unreserve_minor_heaps() {
+static void unreserve_minor_heaps(void) {
   uintnat size;
 
   caml_gc_log("unreserve_minor_heaps");
@@ -812,7 +821,7 @@ static void stw_resize_minor_heap_reservation(caml_domain_state* domain,
        and it also updates the reservation boundaries of each domain
        by mutating its [minor_heap_area_start{,_end}] variables.
 
-       Thse variables are synchronized by the fact that we are inside
+       These variables are synchronized by the fact that we are inside
        a STW section: no other domains are running in parallel, and
        the participating domains will synchronize with this write by
        exiting the barrier, before they read those variables in
@@ -931,7 +940,7 @@ static void* backup_thread_func(void* v)
   domain_self = di;
   caml_state = di->state;
 
-  msg = atomic_load_acq (&di->backup_thread_msg);
+  msg = atomic_load_acquire (&di->backup_thread_msg);
   while (msg != BT_TERMINATE) {
     CAMLassert (msg <= BT_TERMINATE);
     switch (msg) {
@@ -951,7 +960,7 @@ static void* backup_thread_func(void* v)
          * Will be woken from caml_leave_blocking_section
          */
         caml_plat_lock(&s->lock);
-        msg = atomic_load_acq (&di->backup_thread_msg);
+        msg = atomic_load_acquire (&di->backup_thread_msg);
         if (msg == BT_IN_BLOCKING_SECTION &&
             !caml_incoming_interrupts_queued())
           caml_plat_wait(&s->cond);
@@ -963,7 +972,7 @@ static void* backup_thread_func(void* v)
          * or domain_terminate
          */
         caml_plat_lock(&di->domain_lock);
-        msg = atomic_load_acq (&di->backup_thread_msg);
+        msg = atomic_load_acquire (&di->backup_thread_msg);
         if (msg == BT_ENTERING_OCAML)
           caml_plat_wait(&di->domain_cond);
         caml_plat_unlock(&di->domain_lock);
@@ -972,11 +981,11 @@ static void* backup_thread_func(void* v)
         cpu_relax();
         break;
     };
-    msg = atomic_load_acq (&di->backup_thread_msg);
+    msg = atomic_load_acquire (&di->backup_thread_msg);
   }
 
   /* doing terminate */
-  atomic_store_rel(&di->backup_thread_msg, BT_INIT);
+  atomic_store_release(&di->backup_thread_msg, BT_INIT);
 
   return 0;
 }
@@ -992,7 +1001,7 @@ static void install_backup_thread (dom_internal* di)
     CAMLassert (di->backup_thread_msg == BT_INIT || /* Using fresh domain */
             di->backup_thread_msg == BT_TERMINATE); /* Reusing domain */
 
-    while (atomic_load_acq(&di->backup_thread_msg) != BT_INIT) {
+    while (atomic_load_acquire(&di->backup_thread_msg) != BT_INIT) {
       /* Give a chance for backup thread on this domain to terminate */
       caml_plat_unlock (&di->domain_lock);
       cpu_relax ();
@@ -1005,7 +1014,7 @@ static void install_backup_thread (dom_internal* di)
     pthread_sigmask(SIG_BLOCK, &mask, &old_mask);
 #endif
 
-    atomic_store_rel(&di->backup_thread_msg, BT_ENTERING_OCAML);
+    atomic_store_release(&di->backup_thread_msg, BT_ENTERING_OCAML);
     err = pthread_create(&di->backup_thread, 0, backup_thread_func, (void*)di);
 
 #ifndef _WIN32
@@ -1046,7 +1055,7 @@ CAMLexport void (*caml_domain_external_interrupt_hook)(void) =
 CAMLexport _Atomic caml_timing_hook caml_domain_terminated_hook =
   (caml_timing_hook)NULL;
 
-static void domain_terminate();
+static void domain_terminate(void);
 
 static void* domain_thread_func(void* v)
 {
@@ -1220,11 +1229,11 @@ void caml_global_barrier_end(barrier_status b)
   uintnat sense = b & BARRIER_SENSE_BIT;
   if (caml_global_barrier_is_final(b)) {
     /* last domain into the barrier, flip sense */
-    atomic_store_rel(&stw_request.barrier, sense ^ BARRIER_SENSE_BIT);
+    atomic_store_release(&stw_request.barrier, sense ^ BARRIER_SENSE_BIT);
   } else {
     /* wait until another domain flips the sense */
     SPIN_WAIT {
-      uintnat barrier = atomic_load_acq(&stw_request.barrier);
+      uintnat barrier = atomic_load_acquire(&stw_request.barrier);
       if ((barrier & BARRIER_SENSE_BIT) != sense) break;
     }
   }
@@ -1252,7 +1261,7 @@ static void decrement_stw_domains_still_processing(void)
   if( am_last ) {
     /* release the STW lock to allow new STW sections */
     caml_plat_lock(&all_domains_lock);
-    atomic_store_rel(&stw_leader, 0);
+    atomic_store_release(&stw_leader, 0);
     caml_plat_broadcast(&all_domains_cond);
     caml_gc_log("clearing stw leader");
     caml_plat_unlock(&all_domains_lock);
@@ -1265,7 +1274,7 @@ static void stw_handler(caml_domain_state* domain)
   CAML_EV_BEGIN(EV_STW_API_BARRIER);
   {
     SPIN_WAIT {
-      if (atomic_load_acq(&stw_request.domains_still_running) == 0)
+      if (atomic_load_acquire(&stw_request.domains_still_running) == 0)
         break;
 
       if (stw_request.enter_spin_callback)
@@ -1356,8 +1365,16 @@ int caml_domain_is_in_stw(void) {
       finishes.
    The same logic would apply for any other situations in which a domain
    wants to join or leave the set of STW participants.
+
+  The explanation above applies if [sync] = 1. When [sync] = 0, no
+  synchronization happens, and we simply run the handler asynchronously on
+  all domains. We still hold the stw_leader field until we know that
+  every domain has run the handler, so another STW section cannot
+  interfere with this one.
+
 */
 int caml_try_run_on_all_domains_with_spin_work(
+  int sync,
   void (*handler)(caml_domain_state*, void*, int, caml_domain_state**),
   void* data,
   void (*leader_setup)(caml_domain_state*),
@@ -1367,7 +1384,7 @@ int caml_try_run_on_all_domains_with_spin_work(
   int i;
   caml_domain_state* domain_state = domain_self->state;
 
-  caml_gc_log("requesting STW");
+  caml_gc_log("requesting STW, sync=%d", sync);
 
   /* Don't touch the lock if there's already a stw leader
      OR we can't get the lock.
@@ -1377,21 +1394,21 @@ int caml_try_run_on_all_domains_with_spin_work(
      situations. Without this read, [stw_leader] would be protected by
      [all_domains_lock] and could be a non-atomic variable.
   */
-  if (atomic_load_acq(&stw_leader) ||
+  if (atomic_load_acquire(&stw_leader) ||
       !caml_plat_try_lock(&all_domains_lock)) {
     caml_handle_incoming_interrupts();
     return 0;
   }
 
   /* see if there is a stw_leader already */
-  if (atomic_load_acq(&stw_leader)) {
+  if (atomic_load_acquire(&stw_leader)) {
     caml_plat_unlock(&all_domains_lock);
     caml_handle_incoming_interrupts();
     return 0;
   }
 
   /* we have the lock and can claim the stw_leader */
-  atomic_store_rel(&stw_leader, (uintnat)domain_self);
+  atomic_store_release(&stw_leader, (uintnat)domain_self);
 
   CAML_EV_BEGIN(EV_STW_LEADER);
   caml_gc_log("causing STW");
@@ -1402,11 +1419,11 @@ int caml_try_run_on_all_domains_with_spin_work(
   stw_request.enter_spin_data = enter_spin_data;
   stw_request.callback = handler;
   stw_request.data = data;
-  atomic_store_rel(&stw_request.barrier, 0);
-  atomic_store_rel(&stw_request.domains_still_running, 1);
+  atomic_store_release(&stw_request.barrier, 0);
+  atomic_store_release(&stw_request.domains_still_running, sync);
   stw_request.num_domains = stw_domains.participating_domains;
-  atomic_store_rel(&stw_request.num_domains_still_processing,
-                     stw_domains.participating_domains);
+  atomic_store_release(&stw_request.num_domains_still_processing,
+                   stw_domains.participating_domains);
 
   if( leader_setup ) {
     leader_setup(domain_state);
@@ -1432,6 +1449,7 @@ int caml_try_run_on_all_domains_with_spin_work(
     if (d->state != domain_state) caml_send_interrupt(&d->interruptor);
   }
 
+
   /* Domains now know they are part of the STW.
 
      Note: releasing the lock will not allow new domain to be created
@@ -1454,7 +1472,7 @@ int caml_try_run_on_all_domains_with_spin_work(
   }
 
   /* release from the enter barrier */
-  atomic_store_rel(&stw_request.domains_still_running, 0);
+  atomic_store_release(&stw_request.domains_still_running, 0);
 
   #ifdef DEBUG
   domain_state->inside_stw_handler = 1;
@@ -1481,7 +1499,20 @@ int caml_try_run_on_all_domains(
   void (*leader_setup)(caml_domain_state*))
 {
   return
-      caml_try_run_on_all_domains_with_spin_work(handler,
+      caml_try_run_on_all_domains_with_spin_work(1,
+                                                 handler,
+                                                 data,
+                                                 leader_setup, 0, 0);
+}
+
+int caml_try_run_on_all_domains_async(
+  void (*handler)(caml_domain_state*, void*, int, caml_domain_state**),
+  void* data,
+  void (*leader_setup)(caml_domain_state*))
+{
+  return
+      caml_try_run_on_all_domains_with_spin_work(0,
+                                                 handler,
                                                  data,
                                                  leader_setup, 0, 0);
 }
@@ -1492,43 +1523,128 @@ void caml_interrupt_self(void) {
 
 void caml_reset_young_limit(caml_domain_state * dom_st)
 {
+  CAMLassert ((uintnat)dom_st->young_ptr > (uintnat)dom_st->young_trigger);
   /* An interrupt might have been queued in the meanwhile; this
      achieves the proper synchronisation. */
-  atomic_exchange(&dom_st->young_limit, (uintnat)dom_st->young_start);
+  atomic_exchange(&dom_st->young_limit, (uintnat)dom_st->young_trigger);
   dom_internal * d = &all_domains[dom_st->id];
   if (atomic_load_relaxed(&d->interruptor.interrupt_pending)
       || dom_st->requested_minor_gc
       || dom_st->requested_major_slice
+      || dom_st->major_slice_epoch < atomic_load (&caml_major_slice_epoch)
       || atomic_load_relaxed(&dom_st->requested_external_interrupt)
       || dom_st->action_pending) {
-    atomic_store_rel(&dom_st->young_limit, (uintnat)-1);
+    atomic_store_release(&dom_st->young_limit, (uintnat)-1);
     CAMLassert(caml_check_gc_interrupt(dom_st));
   }
 }
 
+Caml_inline void advance_global_major_slice_epoch (caml_domain_state* d)
+{
+  uintnat old_value;
+
+  CAMLassert (atomic_load (&caml_major_slice_epoch) <=
+              atomic_load (&caml_minor_collections_count));
+
+  old_value = atomic_exchange (&caml_major_slice_epoch,
+                               atomic_load (&caml_minor_collections_count));
+
+  if (old_value != atomic_load (&caml_minor_collections_count)) {
+    /* This domain is the first one to use up half of its minor heap arena
+        in this minor cycle. Trigger major slice on other domains. */
+    if (caml_plat_try_lock(&all_domains_lock)) {
+      /* Note that this interrupt is best-effort. If we get the lock,
+         then interrupt all the domains. If not, either some other domain
+         is calling for a stop-the-world section interrupting all the
+         domains, or a domain is being created or terminated. All of these
+         actions also try to lock [all_domains_lock] mutex, and the above
+         lock acquisition may fail.
+
+         If we don't get the lock, we don't interrupt other domains. This is
+         acceptable since it does not affect safety but only liveness -- the
+         speed of the major gc. The other domains may themselves fill half of
+         their minor heap triggering a major slice, or will certainly do a
+         major slice right after their next minor GC when they observe that
+         their domain-local [Caml_state->major_slice_epoch] is less than the
+         global one [caml_major_slice_epoch]. */
+      for(int i = 0; i < stw_domains.participating_domains; i++) {
+        dom_internal * di = stw_domains.domains[i];
+        if (di->state != d) interrupt_domain(&di->interruptor);
+      }
+      caml_plat_unlock (&all_domains_lock);
+    }
+  }
+}
+
+static void global_major_slice_callback (caml_domain_state *domain,
+                                         void *unused,
+                                         int participating_count,
+                                         caml_domain_state **participating)
+{
+  domain->requested_major_slice = 1;
+  /* Nothing else to do, as [stw_hander] will call [caml_poll_gc_work]
+     right after the callback. */
+}
+
 void caml_poll_gc_work(void)
 {
   CAMLalloc_point_here;
 
-  if (((uintnat)Caml_state->young_ptr - Bhsize_wosize(Max_young_wosize) <
-       (uintnat)Caml_state->young_start) ||
-      Caml_state->requested_minor_gc) {
+  caml_domain_state* d = Caml_state;
+
+  if ((uintnat)d->young_ptr - Bhsize_wosize(Max_young_wosize) <
+      (uintnat)d->young_trigger) {
+
+    if (d->young_trigger == d->young_start) {
+      /* Trigger minor GC */
+      d->requested_minor_gc = 1;
+    } else {
+      CAMLassert (d->young_trigger ==
+                  d->young_start + (d->young_end - d->young_start) / 2);
+      /* We have used half of our minor heap arena. Request a major slice on
+         this domain. */
+      advance_global_major_slice_epoch (d);
+      /* Advance the [young_trigger] to [young_start] so that the allocation
+         fails when the minor heap is full. */
+      d->young_trigger = d->young_start;
+    }
+  } else if (d->requested_minor_gc) {
+    /* This domain has _not_ used up half of its minor heap arena, but a minor
+       collection has been requested. Schedule a major collection slice so as
+       to not lag behind. */
+    advance_global_major_slice_epoch (d);
+  }
+
+  if (d->major_slice_epoch < atomic_load (&caml_major_slice_epoch)) {
+    d->requested_major_slice = 1;
+  }
+
+  if (d->requested_minor_gc) {
     /* out of minor heap or collection forced */
-    Caml_state->requested_minor_gc = 0;
+    d->requested_minor_gc = 0;
     caml_empty_minor_heaps_once();
   }
 
-  if (Caml_state->requested_major_slice) {
+  if (d->requested_major_slice || d->requested_global_major_slice) {
     CAML_EV_BEGIN(EV_MAJOR);
-    Caml_state->requested_major_slice = 0;
+    d->requested_major_slice = 0;
     caml_major_collection_slice(AUTO_TRIGGERED_MAJOR_SLICE);
     CAML_EV_END(EV_MAJOR);
   }
 
-  if (atomic_load_acq(&Caml_state->requested_external_interrupt)) {
+  if (d->requested_global_major_slice) {
+    if (caml_try_run_on_all_domains_async(
+          &global_major_slice_callback, NULL, NULL)){
+      d->requested_global_major_slice = 0;
+    }
+    /* If caml_try_run_on_all_domains_async fails, we'll try again next time
+       caml_poll_gc_work is called. */
+  }
+
+  if (atomic_load_acquire(&d->requested_external_interrupt)) {
     caml_domain_external_interrupt_hook();
   }
-  caml_reset_young_limit(Caml_state);
+  caml_reset_young_limit(d);
 }
 
 void caml_handle_gc_interrupt(void)
@@ -1547,7 +1663,7 @@ void caml_handle_gc_interrupt(void)
 
 CAMLexport int caml_bt_is_in_blocking_section(void)
 {
-  uintnat status = atomic_load_acq(&domain_self->backup_thread_msg);
+  uintnat status = atomic_load_acquire(&domain_self->backup_thread_msg);
   return status == BT_IN_BLOCKING_SECTION;
 }
 
@@ -1576,7 +1692,7 @@ CAMLexport void caml_bt_enter_ocaml(void)
   CAMLassert(caml_domain_alone() || self->backup_thread_running);
 
   if (self->backup_thread_running) {
-    atomic_store_rel(&self->backup_thread_msg, BT_ENTERING_OCAML);
+    atomic_store_release(&self->backup_thread_msg, BT_ENTERING_OCAML);
   }
 }
 
@@ -1594,7 +1710,7 @@ CAMLexport void caml_bt_exit_ocaml(void)
   CAMLassert(caml_domain_alone() || self->backup_thread_running);
 
   if (self->backup_thread_running) {
-    atomic_store_rel(&self->backup_thread_msg, BT_IN_BLOCKING_SECTION);
+    atomic_store_release(&self->backup_thread_msg, BT_IN_BLOCKING_SECTION);
     /* Wakeup backup thread if it is sleeping */
     caml_plat_signal(&self->domain_cond);
   }
@@ -1611,7 +1727,8 @@ CAMLexport void (*caml_atfork_hook)(void) = caml_atfork_default;
 static void handover_ephemerons(caml_domain_state* domain_state)
 {
   if (domain_state->ephe_info->todo == 0 &&
-      domain_state->ephe_info->live == 0)
+      domain_state->ephe_info->live == 0 &&
+      domain_state->ephe_info->must_sweep_ephe == 0)
     return;
 
   caml_add_to_orphaned_ephe_list(domain_state->ephe_info);
@@ -1669,7 +1786,6 @@ static void domain_terminate (void)
   call_timing_hook(&caml_domain_terminated_hook);
 
   while (!finished) {
-    caml_orphan_allocated_words();
     caml_finish_sweeping();
 
     caml_empty_minor_heaps_once();
@@ -1752,7 +1868,7 @@ static void domain_terminate (void)
   /* signal the domain termination to the backup thread
      NB: for a program with no additional domains, the backup thread
      will not have been started */
-  atomic_store_rel(&domain_self->backup_thread_msg, BT_TERMINATE);
+  atomic_store_release(&domain_self->backup_thread_msg, BT_TERMINATE);
   caml_plat_signal(&domain_self->domain_cond);
   caml_plat_unlock(&domain_self->domain_lock);
 
