@@ -137,7 +137,7 @@ Caml_inline void write_barrier(
           then this is in a remembered set already */
        if (Is_young(old_val)) return;
        /* old is a block and in the major heap */
-       caml_darken(0, old_val, 0);
+       caml_darken(Caml_state, old_val, 0);
      }
      /* this update is creating a new link from major to minor, remember it */
      if (Is_block_and_young(new_val)) {
@@ -152,8 +152,7 @@ CAMLexport CAMLweakdef void caml_modify (volatile value *fp, value val)
 
   /* See Note [MM] above */
   atomic_thread_fence(memory_order_acquire);
-  atomic_store_explicit(&Op_atomic_val((value)fp)[0], val,
-                        memory_order_release);
+  atomic_store_release(&Op_atomic_val((value)fp)[0], val);
 }
 
 /* Dependent memory is all memory blocks allocated out of the heap
@@ -193,10 +192,9 @@ CAMLexport void caml_adjust_gc_speed (mlsize_t res, mlsize_t max)
   if (max == 0) max = 1;
   if (res > max) res = max;
   Caml_state->extra_heap_resources += (double) res / (double) max;
-  if (Caml_state->extra_heap_resources > 1.0){
+  if (Caml_state->extra_heap_resources > 0.2){
     CAML_EV_COUNTER (EV_C_REQUEST_MAJOR_ADJUST_GC_SPEED, 1);
-    Caml_state->extra_heap_resources = 1.0;
-    caml_request_major_slice ();
+    caml_request_major_slice (1);
   }
 }
 
@@ -328,11 +326,13 @@ CAMLexport void caml_set_fields (value obj, value v)
   }
 }
 
-Caml_inline value alloc_shr(mlsize_t wosize, tag_t tag, int noexc)
+Caml_inline value alloc_shr(mlsize_t wosize, tag_t tag, reserved_t reserved,
+                            int noexc)
 {
   Caml_check_caml_state();
   caml_domain_state *dom_st = Caml_state;
-  value *v = caml_shared_try_alloc(dom_st->shared_heap, wosize, tag, 0);
+  value *v = caml_shared_try_alloc(dom_st->shared_heap,
+                                   wosize, tag, reserved, 0);
   if (v == NULL) {
     if (!noexc)
       caml_raise_out_of_memory();
@@ -341,9 +341,9 @@ Caml_inline value alloc_shr(mlsize_t wosize, tag_t tag, int noexc)
   }
 
   dom_st->allocated_words += Whsize_wosize(wosize);
-  if (dom_st->allocated_words > dom_st->minor_heap_wsz) {
+  if (dom_st->allocated_words > dom_st->minor_heap_wsz / 5) {
     CAML_EV_COUNTER (EV_C_REQUEST_MAJOR_ALLOC_SHR, 1);
-    caml_request_major_slice();
+    caml_request_major_slice(1);
   }
 
 #ifdef DEBUG
@@ -358,11 +358,19 @@ Caml_inline value alloc_shr(mlsize_t wosize, tag_t tag, int noexc)
 
 CAMLexport value caml_alloc_shr(mlsize_t wosize, tag_t tag)
 {
-  return alloc_shr(wosize, tag, 0);
+  return alloc_shr(wosize, tag, 0, 0);
 }
 
+CAMLexport value caml_alloc_shr_reserved(mlsize_t wosize,
+                                         tag_t tag,
+                                         reserved_t reserved)
+{
+  return alloc_shr(wosize, tag, reserved, 0);
+}
+
+
 CAMLexport value caml_alloc_shr_noexc(mlsize_t wosize, tag_t tag) {
-  return alloc_shr(wosize, tag, 1);
+  return alloc_shr(wosize, tag, 0, 1);
 }
 
 /* Global memory pool.
