@@ -17,12 +17,18 @@
 
 open Format
 open Misc
-open Longident
 open Types
 open Toploop
 
-(* The standard output formatter *)
-let std_out = std_formatter
+let error_fmt () =
+  if !Sys.interactive then
+    Format.std_formatter
+  else
+    Format.err_formatter
+
+let action_on_suberror b =
+  if not b && not !Sys.interactive then
+    raise (Compenv.Exit_with_status 125)
 
 (* Directive sections (used in #help) *)
 let section_general = "General"
@@ -122,18 +128,23 @@ let _ = add_directive "cd" (Directive_string dir_cd)
       doc = "Change the current working directory.";
     }
 
-let dir_load ppf name = ignore (Topeval.load_file false ppf name)
 
-let _ = add_directive "load" (Directive_string (dir_load std_out))
+let with_error_fmt f x = f (error_fmt ()) x
+
+let dir_load ppf name =
+  action_on_suberror (Topeval.load_file false ppf name)
+
+let _ = add_directive "load" (Directive_string (with_error_fmt dir_load))
     {
       section = section_run;
       doc = "Load in memory a bytecode object, produced by ocamlc.";
     }
 
-let dir_load_rec ppf name = ignore (Topeval.load_file true ppf name)
+let dir_load_rec ppf name =
+  action_on_suberror (Topeval.load_file true ppf name)
 
 let _ = add_directive "load_rec"
-    (Directive_string (dir_load_rec std_out))
+    (Directive_string (with_error_fmt dir_load_rec))
     {
       section = section_run;
       doc = "As #load, but loads dependencies recursively.";
@@ -143,24 +154,27 @@ let load_file = Topeval.load_file false
 
 (* Load commands from a file *)
 
-let dir_use ppf name = ignore(Toploop.use_file ppf name)
-let dir_use_output ppf name = ignore(Toploop.use_output ppf name)
-let dir_mod_use ppf name = ignore(Toploop.mod_use_file ppf name)
+let dir_use ppf name =
+  action_on_suberror (Toploop.use_input ppf (Toploop.File name))
+let dir_use_output ppf name = action_on_suberror (Toploop.use_output ppf name)
+let dir_mod_use ppf name =
+  action_on_suberror (Toploop.mod_use_input ppf (Toploop.File name))
 
-let _ = add_directive "use" (Directive_string (dir_use std_out))
+let _ = add_directive "use" (Directive_string (with_error_fmt dir_use))
     {
       section = section_run;
       doc = "Read, compile and execute source phrases from the given file.";
     }
 
-let _ = add_directive "use_output" (Directive_string (dir_use_output std_out))
+let _ = add_directive "use_output"
+    (Directive_string (with_error_fmt dir_use_output))
     {
       section = section_run;
       doc = "Execute a command and read, compile and execute source phrases \
              from its output.";
     }
 
-let _ = add_directive "mod_use" (Directive_string (dir_mod_use std_out))
+let _ = add_directive "mod_use" (Directive_string (with_error_fmt dir_mod_use))
     {
       section = section_run;
       doc = "Usage is identical to #use but #mod_use \
@@ -169,149 +183,176 @@ let _ = add_directive "mod_use" (Directive_string (dir_mod_use std_out))
 
 (* Install, remove a printer *)
 
+module Printer = struct
+  type kind =
+    | Old of Types.type_expr
+      (* 'a -> unit *)
+    | Simple of Types.type_expr
+      (* Format.formatter -> 'a -> unit *)
+    | Generic of { ty_path: Path.t; arity: int; }
+      (* (formatter -> 'a1 -> unit) ->
+         (formatter -> 'a2 -> unit) ->
+         ... ->
+         (formatter -> 'an -> unit) ->
+         formatter -> ('a1, 'a2, ..., 'an) t -> unit
+      *)
+end
+
 let filter_arrow ty =
   let ty = Ctype.expand_head !toplevel_env ty in
-  match ty.desc with
+  match get_desc ty with
   | Tarrow (lbl, l, r, _) when not (Btype.is_optional lbl) -> Some (l, r)
   | _ -> None
 
-let rec extract_last_arrow desc =
-  match filter_arrow desc with
-  | None -> raise (Ctype.Unify [])
-  | Some (_, r as res) ->
-      try extract_last_arrow r
-      with Ctype.Unify _ -> res
+let extract_last_arrow ty =
+  let rec extract last ty =
+    match filter_arrow ty with
+    | None -> last
+    | Some ((_, rest) as next) -> extract (Some next) rest
+  in extract None ty
+
+let extract_target_type ty =
+  Option.map fst (extract_last_arrow ty)
 
-let extract_target_type ty = fst (extract_last_arrow ty)
 let extract_target_parameters ty =
-  let ty = extract_target_type ty |> Ctype.expand_head !toplevel_env in
-  match ty.desc with
-  | Tconstr (path, (_ :: _ as args), _)
-      when Ctype.all_distinct_vars !toplevel_env args -> Some (path, args)
-  | _ -> None
+  match extract_target_type ty with
+  | None -> None
+  | Some tgt ->
+      let tgt = Ctype.expand_head !toplevel_env tgt in
+      match get_desc tgt with
+      | Tconstr (path, (_ :: _ as args), _)
+        when Ctype.all_distinct_vars !toplevel_env args ->
+          Some (path, args)
+      | _ -> None
 
-type 'a printer_type_new = Format.formatter -> 'a -> unit
-type 'a printer_type_old = 'a -> unit
-
-let printer_type ppf typename =
-  let printer_type =
-    match
-      Env.find_type_by_name
-        (Ldot(Lident "Topdirs", typename)) !toplevel_env
-    with
-    | path, _ -> path
-    | exception Not_found ->
-        fprintf ppf "Cannot find type Topdirs.%s.@." typename;
-        raise Exit
+let match_simple_printer_type desc ~is_old_style =
+  let make_printer_type =
+    if is_old_style
+    then Topprinters.printer_type_old
+    else Topprinters.printer_type_new
   in
-  printer_type
+  match
+    Ctype.with_local_level ~post:Ctype.generalize begin fun () ->
+      let ty_arg = Ctype.newvar() in
+      Ctype.unify !toplevel_env
+        (make_printer_type ty_arg)
+        (Ctype.instance desc.val_type);
+      ty_arg
+    end
+  with
+  | exception Ctype.Unify _ -> None
+  | ty_arg ->
+      if is_old_style
+      then Some (Printer.Old ty_arg)
+      else Some (Printer.Simple ty_arg)
 
-let match_simple_printer_type desc printer_type =
-  Ctype.begin_def();
-  let ty_arg = Ctype.newvar() in
-  Ctype.unify !toplevel_env
-    (Ctype.newconstr printer_type [ty_arg])
-    (Ctype.instance desc.val_type);
-  Ctype.end_def();
-  Ctype.generalize ty_arg;
-  (ty_arg, None)
+let match_generic_printer_type desc ty_path params =
+  let make_printer_type = Topprinters.printer_type_new in
+  match
+    Ctype.with_local_level ~post:(List.iter Ctype.generalize) begin fun () ->
+      let args = List.map (fun _ -> Ctype.newvar ()) params in
+      let ty_target = Ctype.newty (Tconstr (ty_path, args, ref Mnil)) in
+      let printer_args_ty =
+        List.map (fun ty_var -> make_printer_type ty_var) args in
+      let ty_expected =
+        List.fold_right Topprinters.type_arrow
+          printer_args_ty (make_printer_type ty_target) in
+      Ctype.unify !toplevel_env
+        ty_expected
+        (Ctype.instance desc.val_type);
+      args
+    end
+  with
+  | exception Ctype.Unify _ -> None
+  | args ->
+      if Ctype.all_distinct_vars !toplevel_env args
+      then Some ()
+      else None
 
-let match_generic_printer_type desc path args printer_type =
-  Ctype.begin_def();
-  let args = List.map (fun _ -> Ctype.newvar ()) args in
-  let ty_target = Ctype.newty (Tconstr (path, args, ref Mnil)) in
-  let ty_args =
-    List.map (fun ty_var -> Ctype.newconstr printer_type [ty_var]) args in
-  let ty_expected =
-    List.fold_right
-      (fun ty_arg ty -> Ctype.newty (Tarrow (Asttypes.Nolabel, ty_arg, ty,
-                                             Cunknown)))
-      ty_args (Ctype.newconstr printer_type [ty_target]) in
-  Ctype.unify !toplevel_env
-    ty_expected
-    (Ctype.instance desc.val_type);
-  Ctype.end_def();
-  Ctype.generalize ty_expected;
-  if not (Ctype.all_distinct_vars !toplevel_env args) then
-    raise (Ctype.Unify []);
-  (ty_expected, Some (path, ty_args))
+let match_printer_type desc =
+  match match_simple_printer_type desc ~is_old_style:false with
+  | Some _ as res -> res
+  | None ->
+  match match_simple_printer_type desc ~is_old_style:true with
+  | Some _ as res -> res
+  | None ->
+  match extract_target_parameters desc.val_type with
+  | None -> None
+  | Some (ty_path, args) ->
+    match match_generic_printer_type desc ty_path args with
+    | None -> None
+    | Some () ->
+      Some (Printer.Generic { ty_path; arity = List.length args; })
 
-let match_printer_type ppf desc =
-  let printer_type_new = printer_type ppf "printer_type_new" in
-  let printer_type_old = printer_type ppf "printer_type_old" in
-  try
-    (match_simple_printer_type desc printer_type_new, false)
-  with Ctype.Unify _ ->
-    try
-      (match_simple_printer_type desc printer_type_old, true)
-    with Ctype.Unify _ as exn ->
-      match extract_target_parameters desc.val_type with
-      | None -> raise exn
-      | Some (path, args) ->
-          (match_generic_printer_type desc path args printer_type_new,
-           false)
-
-let find_printer_type ppf lid =
+let find_printer lid =
   match Env.find_value_by_name lid !toplevel_env with
-  | (path, desc) -> begin
-    match match_printer_type ppf desc with
-    | (ty_arg, is_old_style) -> (ty_arg, path, is_old_style)
-    | exception Ctype.Unify _ ->
-      fprintf ppf "%a has a wrong type for a printing function.@."
-      Printtyp.longident lid;
-      raise Exit
-  end
   | exception Not_found ->
-      fprintf ppf "Unbound value %a.@." Printtyp.longident lid;
-      raise Exit
+    let report ppf =
+      fprintf ppf "Unbound value %a.@."
+        Printtyp.longident lid
+    in Error report
+  | (path, desc) ->
+    match match_printer_type desc with
+    | None ->
+      let report ppf =
+        fprintf ppf "%a has the wrong type for a printing function.@."
+          Printtyp.longident lid
+      in Error report
+    | Some kind -> Ok (path, kind)
+
+let install_printer_by_kind path kind =
+  let v = eval_value_path !toplevel_env path in
+  match kind with
+  | Printer.Old ty_arg ->
+    install_printer path ty_arg
+      (fun _formatter repr -> Obj.obj v (Obj.obj repr))
+  | Printer.Simple ty_arg ->
+    install_printer path ty_arg
+      (fun formatter repr -> Obj.obj v formatter (Obj.obj repr))
+  | Printer.Generic { ty_path; arity } ->
+     let rec build v = function
+       | 0 ->
+          Zero
+            (fun formatter repr -> Obj.obj v formatter (Obj.obj repr))
+       | n ->
+          Succ
+            (fun fn -> build ((Obj.obj v : _ -> Obj.t) fn) (n - 1)) in
+     install_generic_printer' path ty_path (build v arity)
+
+let remove_installed_printer path =
+  match remove_printer path with
+  | () -> Ok ()
+  | exception Not_found ->
+    let report ppf =
+      fprintf ppf "The printer named %a is not installed.@."
+        Printtyp.path path
+    in Error report
 
 let dir_install_printer ppf lid =
-  try
-    let ((ty_arg, ty), path, is_old_style) =
-      find_printer_type ppf lid in
-    let v = eval_value_path !toplevel_env path in
-    match ty with
-    | None ->
-       let print_function =
-         if is_old_style then
-           (fun _formatter repr -> Obj.obj v (Obj.obj repr))
-         else
-           (fun formatter repr -> Obj.obj v formatter (Obj.obj repr)) in
-       install_printer path ty_arg print_function
-    | Some (ty_path, ty_args) ->
-       let rec build v = function
-         | [] ->
-            let print_function =
-              if is_old_style then
-                (fun _formatter repr -> Obj.obj v (Obj.obj repr))
-              else
-                (fun formatter repr -> Obj.obj v formatter (Obj.obj repr)) in
-            Zero print_function
-         | _ :: args ->
-            Succ
-              (fun fn -> build ((Obj.obj v : _ -> Obj.t) fn) args) in
-       install_generic_printer' path ty_path (build v ty_args)
-  with Exit -> ()
+  match find_printer lid with
+  | Error report ->
+    report ppf
+  | Ok (path, kind) ->
+    install_printer_by_kind path kind
 
 let dir_remove_printer ppf lid =
-  try
-    let (_ty_arg, path, _is_old_style) = find_printer_type ppf lid in
-    begin try
-      remove_printer path
-    with Not_found ->
-      fprintf ppf "No printer named %a.@." Printtyp.longident lid
-    end
-  with Exit -> ()
+  match find_printer lid with
+  | Error report ->
+    report ppf
+  | Ok (path, _kind) ->
+    match remove_installed_printer path with
+    | Ok () -> ()
+    | Error report -> report ppf
 
 let _ = add_directive "install_printer"
-    (Directive_ident (dir_install_printer std_out))
+    (Directive_ident (with_error_fmt dir_install_printer))
     {
       section = section_print;
       doc = "Registers a printer for values of a certain type.";
     }
 
 let _ = add_directive "remove_printer"
-    (Directive_ident (dir_remove_printer std_out))
+    (Directive_ident (with_error_fmt dir_remove_printer))
     {
       section = section_print;
       doc = "Remove the named function from the table of toplevel printers.";
@@ -319,7 +360,7 @@ let _ = add_directive "remove_printer"
 
 let parse_warnings ppf iserr s =
   try Option.iter Location.(prerr_alert none) @@ Warnings.parse_options iserr s
-  with Arg.Bad err -> fprintf ppf "%s.@." err
+  with Arg.Bad err -> fprintf ppf "%s.@." err; action_on_suberror true
 
 (* Typing information *)
 
@@ -370,7 +411,7 @@ let reg_show_prim name to_sig doc =
   all_show_funs := to_sig :: !all_show_funs;
   add_directive
     name
-    (Directive_ident (show_prim to_sig std_out))
+    (Directive_ident (show_prim to_sig std_formatter))
     {
       section = section_env;
       doc;
@@ -441,11 +482,7 @@ let () =
        let desc = Env.lookup_constructor ~loc Env.Positive lid env in
        if is_exception_constructor env desc.cstr_res then
          raise Not_found;
-       let path =
-         match Ctype.repr desc.cstr_res with
-         | {desc=Tconstr(path, _, _)} -> path
-         | _ -> raise Not_found
-       in
+       let path = Btype.cstr_type_path desc in
        let type_decl = Env.find_type path env in
        if is_extension_constructor desc.cstr_tag then
          let ret_type =
@@ -513,6 +550,9 @@ let is_rec_module id md =
   Btype.unmark_iterators.it_module_declaration Btype.unmark_iterators md;
   rs
 
+let secretly_the_same_path env path1 path2 =
+  let norm path = Printtyp.rewrite_double_underscore_paths env path in
+  Path.same (norm path1) (norm path2)
 
 let () =
   reg_show_prim "show_module"
@@ -522,35 +562,60 @@ let () =
          | Pident id -> id
          | _ -> id
        in
-       let rec accum_aliases md acc =
-         let acc rs =
+       let rec accum_aliases path md acc =
+         let def rs =
            Sig_module (id, Mp_present,
                        {md with md_type = trim_signature md.md_type},
-                       rs, Exported) :: acc in
+                       rs, Exported) in
          match md.md_type with
-         | Mty_alias path ->
-             let md = Env.find_module path env in
-             accum_aliases md (acc Trec_not)
+         | Mty_alias new_path ->
+             let md = Env.find_module new_path env in
+             accum_aliases new_path md
+               (if secretly_the_same_path env path new_path
+                then acc
+                else def Trec_not :: acc)
          | Mty_ident _ | Mty_signature _ | Mty_functor _ ->
-             List.rev (acc (is_rec_module id md))
+             List.rev (def (is_rec_module id md) :: acc)
        in
-       accum_aliases md []
+       accum_aliases path md []
     )
     "Print the signature of the corresponding module."
 
 let () =
   reg_show_prim "show_module_type"
     (fun env loc id lid ->
-       let _path, desc = Env.lookup_modtype ~loc lid env in
-       [ Sig_modtype (id, desc, Exported) ]
+       let path, mtd = Env.lookup_modtype ~loc lid env in
+       let id = match path with
+         | Pident id -> id
+         | _ -> id
+       in
+       let rec accum_defs path mtd acc =
+         let def = Sig_modtype (id, mtd, Exported) in
+         match mtd.mtd_type with
+         | Some (Mty_ident new_path) ->
+             let mtd = Env.find_modtype new_path env in
+             accum_defs new_path mtd
+               (if secretly_the_same_path env path new_path
+                then acc
+                else def :: acc)
+         | None | Some (Mty_alias _ | Mty_signature _ | Mty_functor _) ->
+             List.rev (def :: acc)
+       in
+       accum_defs path mtd []
     )
     "Print the signature of the corresponding module type."
 
 let () =
   reg_show_prim "show_class"
     (fun env loc id lid ->
-       let _path, desc = Env.lookup_class ~loc lid env in
-       [ Sig_class (id, desc, Trec_not, Exported) ]
+       let _path, desc_class = Env.lookup_class ~loc lid env in
+       let _path, desc_cltype = Env.lookup_cltype ~loc lid env in
+       let _path, typedcl = Env.lookup_type ~loc lid env in
+       [
+         Sig_class (id, desc_class, Trec_not, Exported);
+         Sig_class_type (id, desc_cltype, Trec_not, Exported);
+         Sig_type (id, typedcl, Trec_not, Exported);
+       ]
     )
     "Print the signature of the corresponding class."
 
@@ -558,7 +623,11 @@ let () =
   reg_show_prim "show_class_type"
     (fun env loc id lid ->
        let _path, desc = Env.lookup_cltype ~loc lid env in
-       [ Sig_class_type (id, desc, Trec_not, Exported) ]
+       let _path, typedcl = Env.lookup_type ~loc lid env in
+       [
+         Sig_class_type (id, desc, Trec_not, Exported);
+         Sig_type (id, typedcl, Trec_not, Exported);
+       ]
     )
     "Print the signature of the corresponding class type."
 
@@ -571,7 +640,7 @@ let show env loc id lid =
   if sg = [] then raise Not_found else sg
 
 let () =
-  add_directive "show" (Directive_ident (show_prim show std_out))
+  add_directive "show" (Directive_ident (show_prim show std_formatter))
     {
       section = section_env;
       doc = "Print the signatures of components \
@@ -596,6 +665,13 @@ let _ = add_directive "print_length"
 
 (* Set various compiler flags *)
 
+let _ = add_directive "debug"
+    (Directive_bool(fun b -> Clflags.debug := b))
+    {
+      section = section_options;
+      doc = "Choose whether to generate debugging events.";
+    }
+
 let _ = add_directive "labels"
     (Directive_bool(fun b -> Clflags.classic := not b))
     {
@@ -626,14 +702,14 @@ let _ = add_directive "ppx"
     }
 
 let _ = add_directive "warnings"
-    (Directive_string (parse_warnings std_out false))
+    (Directive_string (with_error_fmt(fun ppf s -> parse_warnings ppf false s)))
     {
       section = section_options;
       doc = "Enable or disable warnings according to the argument.";
     }
 
 let _ = add_directive "warn_error"
-    (Directive_string (parse_warnings std_out true))
+    (Directive_string (with_error_fmt(fun ppf s -> parse_warnings ppf true s)))
     {
       section = section_options;
       doc = "Treat as errors the warnings enabled by the argument.";
@@ -703,7 +779,7 @@ let print_directives ppf () =
   List.iter (print_section ppf) (directive_sections ())
 
 let _ = add_directive "help"
-    (Directive_none (print_directives std_out))
+    (Directive_none (print_directives std_formatter))
     {
       section = section_general;
       doc = "Prints a list of all available directives, with \
